@@ -8,17 +8,17 @@ from django.urls import NoReverseMatch, reverse
 from django.utils.safestring import mark_safe
 from django.utils.translation import gettext, gettext_lazy as _, pgettext
 
+from allauth.account.app_settings import LoginMethod
 from allauth.account.internal import flows
+from allauth.account.internal.stagekit import LOGIN_SESSION_KEY
 from allauth.account.stages import EmailVerificationStage
 from allauth.core import context, ratelimit
 from allauth.utils import get_username_max_length, set_form_field_order
 
 from . import app_settings
-from .adapter import get_adapter
-from .app_settings import AuthenticationMethod
+from .adapter import DefaultAccountAdapter, get_adapter
 from .models import EmailAddress, Login
 from .utils import (
-    assess_unique_email,
     filter_users_by_email,
     setup_user_email,
     sync_user_email_addresses,
@@ -47,7 +47,7 @@ class EmailAwarePasswordResetTokenGenerator(PasswordResetTokenGenerator):
 default_token_generator = app_settings.PASSWORD_RESET_TOKEN_GENERATOR()
 
 
-class PasswordVerificationMixin(object):
+class PasswordVerificationMixin:
     def clean(self):
         cleaned_data = super(PasswordVerificationMixin, self).clean()
         password1 = cleaned_data.get("password1")
@@ -96,7 +96,7 @@ class LoginForm(forms.Form):
     def __init__(self, *args, **kwargs):
         self.request = kwargs.pop("request", None)
         super(LoginForm, self).__init__(*args, **kwargs)
-        if app_settings.AUTHENTICATION_METHOD == AuthenticationMethod.EMAIL:
+        if app_settings.LOGIN_METHODS == {LoginMethod.EMAIL}:
             login_widget = forms.EmailInput(
                 attrs={
                     "placeholder": _("Email address"),
@@ -104,7 +104,7 @@ class LoginForm(forms.Form):
                 }
             )
             login_field = forms.EmailField(label=_("Email"), widget=login_widget)
-        elif app_settings.AUTHENTICATION_METHOD == AuthenticationMethod.USERNAME:
+        elif app_settings.LOGIN_METHODS == {LoginMethod.USERNAME}:
             login_widget = forms.TextInput(
                 attrs={"placeholder": _("Username"), "autocomplete": "username"}
             )
@@ -114,10 +114,10 @@ class LoginForm(forms.Form):
                 max_length=get_username_max_length(),
             )
         else:
-            assert (
-                app_settings.AUTHENTICATION_METHOD
-                == AuthenticationMethod.USERNAME_EMAIL
-            )
+            assert app_settings.LOGIN_METHODS == {
+                LoginMethod.USERNAME,
+                LoginMethod.EMAIL,
+            }  # nosec
             login_widget = forms.TextInput(
                 attrs={"placeholder": _("Username or email"), "autocomplete": "email"}
             )
@@ -136,7 +136,7 @@ class LoginForm(forms.Form):
             forgot_txt = _("Forgot your password?")
             self.fields["password"].help_text = mark_safe(
                 f'<a href="{reset_url}">{forgot_txt}</a>'
-            )
+            )  # nosec
 
     def user_credentials(self):
         """
@@ -145,14 +145,13 @@ class LoginForm(forms.Form):
         """
         credentials = {}
         login = self.cleaned_data["login"]
-        if app_settings.AUTHENTICATION_METHOD == AuthenticationMethod.EMAIL:
+        method = flows.login.derive_login_method(login)
+        if method == LoginMethod.EMAIL:
             credentials["email"] = login
-        elif app_settings.AUTHENTICATION_METHOD == AuthenticationMethod.USERNAME:
+        elif method == LoginMethod.USERNAME:
             credentials["username"] = login
         else:
-            if self._is_login_email(login):
-                credentials["email"] = login
-            credentials["username"] = login
+            raise NotImplementedError()
         credentials["password"] = self.cleaned_data["password"]
         return credentials
 
@@ -160,14 +159,6 @@ class LoginForm(forms.Form):
         login = self.cleaned_data["login"]
         return login.strip()
 
-    def _is_login_email(self, login):
-        try:
-            validators.validate_email(login)
-            ret = True
-        except exceptions.ValidationError:
-            ret = False
-        return ret
-
     def clean(self):
         super(LoginForm, self).clean()
         if self._errors:
@@ -176,25 +167,22 @@ class LoginForm(forms.Form):
         adapter = get_adapter(self.request)
         user = adapter.authenticate(self.request, **credentials)
         if user:
+            login = Login(user=user, email=credentials.get("email"))
+            if flows.login.is_login_rate_limited(context.request, login):
+                raise adapter.validation_error("too_many_login_attempts")
+            self._login = login
             self.user = user
         else:
-            auth_method = app_settings.AUTHENTICATION_METHOD
-            if auth_method == app_settings.AuthenticationMethod.USERNAME_EMAIL:
-                login = self.cleaned_data["login"]
-                if self._is_login_email(login):
-                    auth_method = app_settings.AuthenticationMethod.EMAIL
-                else:
-                    auth_method = app_settings.AuthenticationMethod.USERNAME
-            raise adapter.validation_error("%s_password_mismatch" % auth_method)
+            login_method = flows.login.derive_login_method(
+                login=self.cleaned_data["login"]
+            )
+            raise adapter.validation_error("%s_password_mismatch" % login_method.value)
         return self.cleaned_data
 
     def login(self, request, redirect_url=None):
         credentials = self.user_credentials()
-        login = Login(
-            user=self.user,
-            redirect_url=redirect_url,
-            email=credentials.get("email"),
-        )
+        login = self._login
+        login.redirect_url = redirect_url
         ret = flows.login.perform_password_login(request, credentials, login)
         remember = app_settings.SESSION_REMEMBER
         if remember is None:
@@ -254,7 +242,7 @@ def _base_signup_form_class():
     return fc_class
 
 
-class BaseSignupForm(_base_signup_form_class()):
+class BaseSignupForm(_base_signup_form_class()):  # type: ignore[misc]
     username = forms.CharField(
         label=_("Username"),
         min_length=app_settings.USERNAME_MIN_LENGTH,
@@ -348,7 +336,7 @@ class BaseSignupForm(_base_signup_form_class()):
 
     def validate_unique_email(self, value):
         adapter = get_adapter()
-        assessment = assess_unique_email(value)
+        assessment = flows.manage_email.assess_unique_email(value)
         if assessment is True:
             # All good.
             pass
@@ -356,7 +344,7 @@ class BaseSignupForm(_base_signup_form_class()):
             # Fail right away.
             raise adapter.validation_error("email_taken")
         else:
-            assert assessment is None
+            assert assessment is None  # nosec
             self.account_already_exists = True
         return adapter.validate_unique_email(value)
 
@@ -381,11 +369,10 @@ class BaseSignupForm(_base_signup_form_class()):
             # Don't create a new account, only send an email informing the user
             # that (s)he already has one...
             email = self.cleaned_data["email"]
-            adapter = get_adapter()
-            adapter.send_account_already_exists_mail(email)
+            resp = flows.signup.prevent_enumeration(request, email)
             user = None
-            resp = adapter.respond_email_verification_sent(request, None)
-            request.session[flows.login.LOGIN_SESSION_KEY] = EmailVerificationStage.key
+            # Fake a login stage.
+            request.session[LOGIN_SESSION_KEY] = EmailVerificationStage.key
         else:
             user = self.save(request)
             resp = None
@@ -406,16 +393,18 @@ class BaseSignupForm(_base_signup_form_class()):
 
 class SignupForm(BaseSignupForm):
     def __init__(self, *args, **kwargs):
+        self.by_passkey = kwargs.pop("by_passkey", False)
         super(SignupForm, self).__init__(*args, **kwargs)
-        self.fields["password1"] = PasswordField(
-            label=_("Password"),
-            autocomplete="new-password",
-            help_text=password_validation.password_validators_help_text_html(),
-        )
-        if app_settings.SIGNUP_PASSWORD_ENTER_TWICE:
-            self.fields["password2"] = PasswordField(
-                label=_("Password (again)"), autocomplete="new-password"
+        if not self.by_passkey:
+            self.fields["password1"] = PasswordField(
+                label=_("Password"),
+                autocomplete="new-password",
+                help_text=password_validation.password_validators_help_text_html(),
             )
+            if app_settings.SIGNUP_PASSWORD_ENTER_TWICE:
+                self.fields["password2"] = PasswordField(
+                    label=_("Password (again)"), autocomplete="new-password"
+                )
 
         if hasattr(self, "field_order"):
             set_form_field_order(self, self.field_order)
@@ -511,9 +500,12 @@ class AddEmailForm(UserForm):
         if on_this_account:
             raise adapter.validation_error("duplicate_email")
         if (
+            # Email is taken by a different account...
             on_diff_account
-            and app_settings.PREVENT_ENUMERATION != "strict"
+            # We care about not having duplicate emails
             and app_settings.UNIQUE_EMAIL
+            # Enumeration prevention is turned off.
+            and (not app_settings.PREVENT_ENUMERATION)
         ):
             raise adapter.validation_error("email_taken")
         if not EmailAddress.objects.can_add_email(self.user):
@@ -529,7 +521,13 @@ class AddEmailForm(UserForm):
         return value
 
     def save(self, request):
-        if app_settings.CHANGE_EMAIL:
+        if app_settings.EMAIL_VERIFICATION_BY_CODE_ENABLED:
+            email_address = EmailAddress(
+                user=self.user, email=self.cleaned_data["email"]
+            )
+            email_address.send_confirmation(request)
+            return email_address
+        elif app_settings.CHANGE_EMAIL:
             return EmailAddress.objects.add_new_email(
                 request, self.user, self.cleaned_data["email"]
             )
@@ -591,19 +589,15 @@ class ResetPasswordForm(forms.Form):
             raise get_adapter().validation_error("unknown_email")
         return self.cleaned_data["email"]
 
-    def save(self, request, **kwargs):
+    def save(self, request, **kwargs) -> str:
         email = self.cleaned_data["email"]
         if not self.users:
-            if app_settings.EMAIL_UNKNOWN_ACCOUNTS:
-                flows.signup.send_unknown_account_mail(request, email)
-        else:
-            self._send_password_reset_mail(request, email, self.users, **kwargs)
-        return email
+            flows.signup.send_unknown_account_mail(request, email)
+            return email
 
-    def _send_password_reset_mail(self, request, email, users, **kwargs):
+        adapter: DefaultAccountAdapter = get_adapter()
         token_generator = kwargs.get("token_generator", default_token_generator)
-
-        for user in users:
+        for user in self.users:
             temp_key = token_generator.make_token(user)
 
             # send the password reset email
@@ -612,7 +606,7 @@ class ResetPasswordForm(forms.Form):
             # not implementation details such as a separate `uidb36` and
             # `key. Ideally, this should have done on `urls` level as well.
             key = f"{uid}-{temp_key}"
-            url = get_adapter().get_reset_password_from_key_url(key)
+            url = adapter.get_reset_password_from_key_url(key)
             context = {
                 "user": user,
                 "password_reset_url": url,
@@ -621,9 +615,10 @@ class ResetPasswordForm(forms.Form):
                 "request": request,
             }
 
-            if app_settings.AUTHENTICATION_METHOD != AuthenticationMethod.EMAIL:
+            if LoginMethod.USERNAME in app_settings.LOGIN_METHODS:
                 context["username"] = user_username(user)
-            get_adapter().send_mail("account/email/password_reset_key", email, context)
+            adapter.send_password_reset_mail(user, email, context)
+        return email
 
 
 class ResetPasswordKeyForm(PasswordVerificationMixin, forms.Form):
@@ -712,7 +707,7 @@ class RequestLoginCodeForm(forms.Form):
         return email
 
 
-class ConfirmLoginCodeForm(forms.Form):
+class BaseConfirmCodeForm(forms.Form):
     code = forms.CharField(
         label=_("Code"),
         widget=forms.TextInput(
@@ -725,8 +720,15 @@ class ConfirmLoginCodeForm(forms.Form):
         super().__init__(*args, **kwargs)
 
     def clean_code(self):
-        code = self.cleaned_data.get("code").replace(" ", "").lower()
-        expected_code = self.code.replace(" ", "").lower()
-        if not self.code or code != expected_code:
+        code = self.cleaned_data.get("code")
+        if not flows.login_by_code.compare_code(actual=code, expected=self.code):
             raise get_adapter().validation_error("incorrect_code")
         return code
+
+
+class ConfirmLoginCodeForm(BaseConfirmCodeForm):
+    pass
+
+
+class ConfirmEmailVerificationCodeForm(BaseConfirmCodeForm):
+    pass
