@@ -1,5 +1,7 @@
+from django.test import TestCase
+
 from allauth.socialaccount.tests import OAuth2TestsMixin
-from allauth.tests import MockedResponse, TestCase
+from allauth.tests import MockedResponse
 
 from .provider import AsanaProvider
 
@@ -15,3 +17,6 @@ class AsanaTests(OAuth2TestsMixin, TestCase):
 {"id": 3133777, "name": "Personal Projects"}], "email": "test@example.com",
 "name": "Test Name", "id": 43748387}}""",
         )
+
+    def get_expected_to_str(self):
+        return "test@example.com"
